package com.bergerkiller.bukkit.nolagg;

import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;

import org.bukkit.Bukkit;
import org.bukkit.ChatColor;
import org.bukkit.Chunk;
import org.bukkit.Material;
import org.bukkit.World;
import org.bukkit.command.CommandSender;
import org.bukkit.entity.Entity;
import org.bukkit.entity.ExperienceOrb;
import org.bukkit.entity.Item;
import org.bukkit.entity.Player;
import org.bukkit.entity.TNTPrimed;
import org.bukkit.plugin.Plugin;

import com.all.bukkit.threadcheck.ThreadCheck;
import com.bergerkiller.bukkit.common.config.ConfigurationNode;
import com.bergerkiller.bukkit.common.config.FileConfiguration;
import com.bergerkiller.bukkit.common.PluginBase;
import com.bergerkiller.bukkit.common.Task;
import com.bergerkiller.bukkit.common.permissions.NoPermissionException;
import com.bergerkiller.bukkit.common.utils.CommonUtil;
import com.bergerkiller.bukkit.common.utils.EntityUtil;
import com.bergerkiller.bukkit.common.utils.EnumUtil;
import com.bergerkiller.bukkit.nolagg.datapool.Pool;
import com.bergerkiller.bukkit.nolagg.logging.PluginLogger;
import com.bergerkiller.bukkit.nolagg.logging.SchedulerWatcher;
import com.bergerkiller.bukkit.nolagg.meta.ChunkMetaData;
import com.bergerkiller.bukkit.nolagg.meta.WorldMetaData;
import com.bergerkiller.bukkit.nolagg.sending.ChunkSendQueue;
import com.bergerkiller.bukkit.nolagg.sending.ChunkCompressionThread;
import com.bergerkiller.bukkit.nolagg.spawnlimit.SpawnHandler;

public class NoLagg extends PluginBase {

	public NoLagg() {
		super(1938, 1960);
	}

	public static NoLagg plugin;

	private Task updateTask;
	public int updateInterval;

	public static boolean tntChangeBlocks = true;
	public static boolean bufferItems;
	public static boolean bufferTNT;
	public static boolean useSpawnLimits;
	public static boolean isOreObfEnabled = false;
	public static boolean isOreToClientObfEnabled = false;
	public static boolean isSpoutEnabled = false;
	public static boolean autoFixLighting = true;
	public static boolean stackOrbs = true;
	public static boolean ignoreMobLimits = false;

	private static Map<String, List<String>> clearShortcuts = new HashMap<String, List<String>>();

	@Override
	public void updateDependency(Plugin plugin, String pluginName, boolean enabled) {
		if (pluginName.equals("Spout")) {
			isSpoutEnabled = enabled;
		} else if (pluginName.equals("Chunk Manager")) {
			if (ChunkSendQueue.enabled && enabled) {
				log(Level.WARNING, "Chunk Manager detected, internal chunk sending system has been disabled.");
				log(Level.WARNING, "Either disable Chunk Manager or set chunkSending.enabled to false in the configuration.");
				ChunkSendQueue.enabled = false;
			}
		} else if (pluginName.equals("Orebfuscator")) {
			if (isOreObfEnabled = enabled) {
				log(Level.INFO, "Orebfuscation has been detected and will be used when sending chunks");
				log(Level.INFO, "Note that this may require you to set more threads used for sending!");
			}
		} else if (pluginName.equals("RawcriticsOreObfuscationPluginSpout")) {
			if (isOreToClientObfEnabled = enabled) {
				log(Level.INFO, "Ore To Client Obfuscation (Spout) has been detected and will be used when sending chunks");
				log(Level.INFO, "Note that this may require you to set more threads used for sending!");
			}
		} else if (pluginName.equals("EntitySuppresor")) {
			if (ignoreMobLimits = enabled) {
				log(Level.WARNING, "Entity suppresor detected: Mobs will not be limited!");
			}
		}
	}

	public void enable() {			
		ThreadCheck.init(this);
		plugin = this;
		SchedulerWatcher.init();

		//General registering
		this.register(NLListener.class);
		this.registerNextTick(NLMonitoredListener.class);
		this.register("nolagg");				

		this.loadConfig();
		Pool.init();
		StackFormer.init();
		TnTHandler.init();
		this.updateTask = new Task(this) {
			public void run() {
				try {				
					WorldMetaData.updateAll();
				} catch (Throwable t) {
					log(Level.SEVERE, "An error occured while performing a routine update:");
					t.printStackTrace();
				}
			}
		}.start(0, updateInterval);

		OperationScheduler.init();
		AutoSaveChanger.init();
		ChunkSendQueue.init();
	}
	public void disable() {
		SchedulerWatcher.deinit();
		PluginLogger.stopTask();
		ThreadCheck.deinit();
		Task.stop(this.updateTask);
		TnTHandler.deinit();
		PerformanceMonitor.deinit();
		OperationScheduler.deinit();
		ChunkSendQueue.deinit();
		SpawnHandler.clear();
		AutoSaveChanger.deinit();
		WorldMetaData.deinit();
		StackFormer.deinit();
		Pool.deinit();
		ChunkCompressionThread.deinit();
	}

	public void loadConfig() {
		//General settings
		FileConfiguration config = new FileConfiguration(this);
		boolean regen = !config.exists();
		config.load();

		//Top header
		config.setHeader("This is the configuration of NoLagg, in here you can enable or disable features as you please");
		config.addHeader("For more information, you can visit the following websites:");
		config.addHeader("http://dev.bukkit.org/server-mods/nolagg/");
		config.addHeader("http://forums.bukkit.org/threads/nolagg.36986/");

		//unload delay
		config.setHeader("chunkUnloadDelay", "");
		config.setHeader("chunkUnloadDelay.enabled", "Whether or not chunks are unloaded after a set delay");
		config.setHeader("chunkUnloadDelay.delay", "The unload delay in milliseconds");
		ChunkMetaData.useUnloadDelay = config.get("chunkUnloadDelay.enabled", true);
		ChunkMetaData.unloadDelay = config.get("chunkUnloadDelay.delay", 10000);

		//tnt
		config.setHeader("TNTHandler", "");
		config.setHeader("TNTHandler.enabled", "Whether or not the TNT feature is used");
		config.setHeader("TNTHandler.detonationInterval", "The interval (in ticks) at which TNT is detonated by explosions");
		config.setHeader("TNTHandler.detonationRate", "How many TNT is detonated by explosions per interval");
		config.setHeader("TNTHandler.ExplosionRadiusFactor", "The explosion crater size factor");
		config.setHeader("TNTHandler.explosionRate", "The amount of explosion packets to send to the clients per tick");
		config.setHeader("TNTHandler.changeBlocks", "If TNT explosions can change non-TNT blocks");
		bufferTNT = config.get("TNTHandler.enabled", true);
		TnTHandler.interval = config.get("TNTHandler.detonationInterval", 1);
		TnTHandler.rate = config.get("TNTHandler.detonationRate", 10);
		CustomExplosion.factor = config.get("TNTHandler.ExplosionRadiusFactor", 1.0);
		TnTHandler.explosionRate = config.get("TNTHandler.explosionRate", 5);
		tntChangeBlocks = config.get("TNTHandler.changeBlocks", true);

		//item buffering	
		config.setHeader("itemBuffering", "");
		config.setHeader("itemBuffering.enabled", "Whether or not items are buffered (spawned at a later time) in chunks");
		config.setHeader("itemBuffering.maxPerChunk", "The maximum amount of items allowed per chunk");
		bufferItems = config.get("itemBuffering.enabled", true);
		ItemHandler.maxItemsPerChunk = config.get("itemBuffering.maxPerChunk", 40);

		//item stacking
		config.setHeader("itemStacking", "");
		config.setHeader("itemStacking.enabled", "Whether or not items are stacked to one item");
		config.setHeader("itemStacking.radius", "The block radius to look for other items when stacking");
		config.setHeader("itemStacking.threshold", "The amount of (physical) items needed to form one stack");
		ItemHandler.formStacks = config.get("itemStacking.enabled", true);
		StackFormer.stackRadiusSquared = Math.pow(config.get("itemStacking.radius", 1.0), 2);
		StackFormer.stackThreshold = config.get("itemStacking.threshold", 2);

		//ignored types
		if (regen) {
			config.set("ignoredItemTypes", Arrays.asList("DIAMOND_PICKAXE", "WOODEN_HOE"));
		}
		config.setHeader("ignoredItemTypes", "");
		config.addHeader("ignoredItemTypes", "The item types (materials) to ignore during item stacking, buffering and spawn limits");
		config.addHeader("ignoredItemTypes", "Use 'orb' to ignore experience orbs");
		ItemHandler.clearIgnoredTypes();
		stackOrbs = true;
		for (String type : config.getList("ignoredItemTypes", String.class)) {
			if (type.equalsIgnoreCase("experienceorb") || type.equalsIgnoreCase("xporb") || type.equalsIgnoreCase("orb")) {
				stackOrbs = false;
				continue;
			}
			Material mat = EnumUtil.parseMaterial(type, null);
			if (mat != null) {
				ItemHandler.addIgnoredType(mat);
			} else {
				log(Level.WARNING, "Unknown item type found in configuration: " + type);
			}
		}

		//general intervals and settings
		config.setHeader("autoSaveInterval", "");
		config.addHeader("autoSaveInterval", "The tick interval at which the server saves automatically (20 ticks = 1 second)");
		config.setHeader("updateInterval", "The tick interval at which NoLagg updates internally (item stacking, buffering and chunk unloading) (20 ticks = 1 second)");
		config.setHeader("monitorInterval", "The tick interval at which performance information is sent and/or logged (20 ticks = 1 second)");
		config.setHeader("startLoggingOnStartup", "Whether or not to start logging to the log.txt file when NoLagg enables");
		config.setHeader("autoFixLighting", "Whether or not the lighting of newly-generated chunks is fixed");
		AutoSaveChanger.newInterval = config.get("autoSaveInterval", 400);
		updateInterval = config.get("updateInterval", 20);
		PerformanceMonitor.deinit();
		PerformanceMonitor.monitorInterval = config.get("monitorInterval", 40);
		PerformanceMonitor.sendLog = config.get("startLoggingOnStartup", false);
		PerformanceMonitor.init();
		autoFixLighting = config.get("autoFixLighting", true);

		//chunk sending
		config.setHeader("chunkSending", "");
		config.setHeader("chunkSending.enabled", "Whether or not the alternative chunk sending alghorithms are used");
		config.setHeader("chunkSending.minRate", "The minimum chunk sending rate (chunks/tick)");
		config.setHeader("chunkSending.maxRate", "The maximum chunk sending rate (chunks/tick)");
		config.setHeader("chunkSending.triggerRate", "The rate at which chunks are 'triggered' (refreshed, make visible) (trigger packets/tick)");
		config.setHeader("chunkSending.sendThreadCount", "The amount of threads to use to compress the chunk packets (increase if it can't keep up)");
		ChunkSendQueue.enabled = config.get("chunkSending.enabled", true);
		ChunkSendQueue.minRate = config.get("chunkSending.minRate", 0.25);
		ChunkSendQueue.maxRate = config.get("chunkSending.maxRate", 4.00);
		ChunkSendQueue.globalTriggerRate = config.get("chunkSending.triggerRate", 0.5);
		if (ChunkSendQueue.enabled) {
			ChunkCompressionThread.init(config.get("chunkSending.sendThreadCount", 2));
		}

		//clear shortcuts
		if (regen) {
			ConfigurationNode node = config.getNode("clearShortcuts");
			node.set("enemies", Arrays.asList("monster"));
			node.set("notneutral", Arrays.asList("monster", "item", "tnt", "egg", "arrow"));
		}
		ConfigurationNode shortc = config.getNode("clearShortcuts");
		shortc.setHeader("");
		shortc.addHeader("Several shortcuts you can use for the /nolagg clear(all) command");
		for (String key : shortc.getKeys()) {
			clearShortcuts.put(key.toLowerCase(), shortc.getList(key, String.class));
		}

		//spawn limits
		config.setHeader("useEntitySpawnLimits", "");
		config.addHeader("useEntitySpawnLimits", "Whether or not entity spawn limits are used");
		useSpawnLimits = config.get("useEntitySpawnLimits", true);

		if (regen) {
			ConfigurationNode limits = config.getNode("spawnlimits");
			ConfigurationNode node = limits.getNode("default");
			node.set("mob", 800);
			node.set("cow", 60);
			node.set("item", 600);
			node.set("creepers", 30);
			node.set("monsters", 400);
			node = limits.getNode("worlds");
			node.set("world1.monsters", 300);
			node.set("world2.chickens", 30);
			node = limits.getNode("global");
			node.set("mobs", 3000);
			limits = config.getNode("mobSpawnerLimits");
			limits.set("default.mob", 300);
			limits.set("default.zombie", 30);
			limits.set("worlds.world3.cavespider", 0);
			limits.set("worlds.creativeworld.mob", 0);
			limits.set("global.mob", 1000);
			//others
		}

		config.setHeader("spawnlimits", "");
		config.addHeader("spawnlimits", "The general spawn limits (natural spawning)");
		config.addHeader("mobSpawnlimits", "For more information, see http://dev.bukkit.org/server-mods/nolagg/pages/spawn-limits-nolagg/");
		config.setHeader("spawnlimits.default", "The default spawn limits per world, overridden by world limits");
		config.setHeader("spawnlimits.worlds", "The world-specific spawn limits");
		config.setHeader("spawnlimits.global", "The global spawn limits");

		config.setHeader("mobSpawnerLimits", "");
		config.addHeader("mobSpawnerLimits", "The spawn limits for mob spawners");
		config.addHeader("mobSpawnerLimits", "For more information, see http://dev.bukkit.org/server-mods/nolagg/pages/spawn-limits-nolagg/");
		config.setHeader("mobSpawnerLimits.default", "The default spawn limits per world, overridden by world limits");
		config.setHeader("mobSpawnerLimits.worlds", "The world-specific spawn limits");
		config.setHeader("mobSpawnerLimits.global", "The global spawn limits");

		SpawnHandler.clear();
		if (useSpawnLimits) {
			//Spawn restrictions
			SpawnHandler.GENERALHANDLER.load(config.getNode("spawnlimits"));
			//Mob spawn restrictions
			SpawnHandler.MOBSPAWNERHANDLER.load(config.getNode("mobSpawnerLimits"));
		}

		//Write out data
		config.save(); 
	}

	private String[] lastargs = new String[1];

	/*
	 * Note to self: this heavily needs improvements in Permission handling!
	 */
	@Override
	public boolean command(CommandSender sender, String cmdLabel, String[] args) {
		try {
			if (args.length > 0) {
				boolean all = args[0].equalsIgnoreCase("clearall");
				if (args[0].equalsIgnoreCase("clear") || all) {
					if (sender instanceof Player) {
						Permission.CLEAR.handle(sender);
					} else {
						all = true;
					}
					//fix and partly read args
					boolean tnt = args.length == 1;
					boolean items = tnt;
					boolean animals = false;
					boolean monsters = false;
					boolean remall = false;
					if (args.length == 2) {
						if (args[1].equalsIgnoreCase("last")) {
							args = lastargs;
							sender.sendMessage(ChatColor.GREEN + "The last-used clear command has been invoked:");
						} else {
							List<String> a = clearShortcuts.get(args[1].toLowerCase());
							if (a != null) {
								args = a.toArray(new String[0]);
								sender.sendMessage(ChatColor.GREEN + "Using clear shortcut: " + args[1]);
							}
						}
					}
					String[] toremove = new String[args.length - 1];
					for (int i = 1; i < args.length; i++) {
						String name = args[i].toLowerCase();
						if (name.contains("tnt")) {
							tnt = true;
						} else if (name.contains("item")) {
							items = true;
						} else if (name.contains("animal")) {
							animals = true;
						} else if (name.contains("monster")) {
							monsters = true;
						} else if (name.contains("mob")) {
							animals = true;
							monsters = true;
						} else if (name.contains("all")) {
							remall = true;
						}
						toremove[i - 1] = name;
					}
					args = toremove;
					World[] worlds;
					if (all) {
						worlds = Bukkit.getServer().getWorlds().toArray(new World[0]);
						if (items) WorldMetaData.clearAllItems();
					} else {
						worlds = new World[] {((Player) sender).getWorld()};
						if (items) WorldMetaData.get(worlds[0]).clearItems();
					}
					if (tnt) TnTHandler.clear();
					int remcount = 0;
					for (World world : worlds) {
						for (Entity e : world.getEntities()) {
							boolean remove = false;
							if (e instanceof Player) {
								continue;
							} else if (remall) {
								remove = true;
							} else if (args.length == 0) {
								remove = e instanceof Item || e instanceof TNTPrimed || e instanceof ExperienceOrb;
							} else if (e instanceof TNTPrimed && tnt) {
								remove = true;
							} else if (e instanceof Item && items) {
								remove = true;
							} else {
								String type = EntityUtil.getName(e);
								if (animals && EntityUtil.isAnimal(type)) {
									remove = true;
								} else if (monsters && EntityUtil.isMonster(type)) {
									remove = true;
								} else {
									for (String arg : args) {
										if (type.contains(arg) || arg.contains(type)) {
											remove = true;
											break;
										}
									}	
								}
							}
							if (remove) {
								e.remove();
								remcount++;
							}
						}
					}
					if (all) {
						sender.sendMessage(ChatColor.YELLOW + "All worlds have been cleared: " + remcount + " entities removed!");
					} else {
						sender.sendMessage(ChatColor.YELLOW + "This world has been cleared: " + remcount + " entities removed!");
					}
				} else if (args[0].equalsIgnoreCase("monitor") || args[0].equalsIgnoreCase("stat") || args[0].equalsIgnoreCase("stats")) {
					boolean once = !args[0].equalsIgnoreCase("monitor");
					if (sender instanceof Player) {
						Permission.MONITOR.handle(sender);
						Player p = (Player) sender;
						if (PerformanceMonitor.recipients.remove(p.getName())) {
							for (int i = 0; i < 10; i++) {
								p.sendMessage(" ");
							}
							p.sendMessage(ChatColor.YELLOW + "You are no longer monitoring this server.");
						} else {
							PerformanceMonitor.recipients.add(p.getName());
							if (once) {
								p.sendMessage(ChatColor.GREEN + "You will receive a snapshot shortly...");
								PerformanceMonitor.removalReq.add(p.getName());
							} else {
								p.sendMessage(ChatColor.GREEN + "You are now monitoring this server.");
								p.sendMessage(ChatColor.GREEN + "To stop monitoring, perform this command again.");
								PerformanceMonitor.removalReq.remove(p.getName());
							}
						}
					} else {
						if (PerformanceMonitor.sendConsole) {
							PerformanceMonitor.sendConsole = false;
							sender.sendMessage("You are no longer monitoring this server.");
						} else {
							PerformanceMonitor.sendConsole = true;
							if (once) {
								sender.sendMessage("You will receive a snapshot shortly...");
								PerformanceMonitor.removalCon = true;
							} else {
								sender.sendMessage("You are now monitoring this server.");
								sender.sendMessage("To stop monitoring, perform this command again.");
								PerformanceMonitor.removalCon = false;
							}
						}
					}
				} else if (args[0].equalsIgnoreCase("log")) {
					Permission.LOG.handle(sender);
					if (PerformanceMonitor.sendLog) {
						PerformanceMonitor.sendLog = false;
						sender.sendMessage("Server stats are no longer logged to file.");
					} else {
						PerformanceMonitor.sendLog = true;
						sender.sendMessage("Server stats are now logged to file.");
						sender.sendMessage("To stop logging, perform this command again.");
					}
				} else if (args[0].equalsIgnoreCase("clearlog")) {
					Permission.CLEARLOG.handle(sender);
					if (PerformanceMonitor.clearLog()) {
						sender.sendMessage("Server log cleared");
					} else {
						sender.sendMessage("Failed to clear the server log");
					}
				} else if (args[0].equalsIgnoreCase("reload")) {
					Permission.RELOAD.handle(sender);
					this.loadConfig();
					sender.sendMessage("Configuration of NoLagg reloaded");
				} else if (args[0].equalsIgnoreCase("fix")) {
					if (sender instanceof Player) {
						Permission.FIX.handle(sender);
						Player p = (Player) sender;
						int radius = Bukkit.getServer().getViewDistance();
						if (args.length == 2) {
							try {
								radius = Integer.parseInt(args[1]);
							} catch (Exception ex) {}
						}
						int cx = p.getLocation().getBlockX() >> 4;
							int cz = p.getLocation().getBlockZ() >> 4;
							for (int a = -radius; a <= radius; a++) {
								for (int b = -radius; b <= radius; b++) {
									Chunk c = p.getWorld().getChunkAt(cx + a, cz + b);
									OperationScheduler.start(new ChunkFixLightingOperation(c));
								} 
							}
							p.sendMessage(ChatColor.GREEN + "A " + (radius * 2 + 1) + " X " + (radius * 2 + 1) + " chunk area around you is currently being fixed from lighting issues...");
					} else {
						sender.sendMessage("This command is only for players!");
					}
				} else if (args[0].equalsIgnoreCase("gc")) {
					Permission.GC.handle(sender);
					Runtime.getRuntime().gc();
					sender.sendMessage("Memory garbage collected!");
				} else if (args[0].equalsIgnoreCase("mem") || args[0].equalsIgnoreCase("memory")) {
					if (sender instanceof Player) {
						Permission.SHOWMEMORY.handle(sender);
						sender.sendMessage(ChatColor.GREEN + "[Static memory] " + ChatColor.YELLOW + "[Dynamic memory] " + ChatColor.RED + "[Unused memory]");
					}
					sender.sendMessage(PerformanceMonitor.getMemory(sender instanceof Player));
				} else if (args[0].equalsIgnoreCase("lagmem") || args[0].equalsIgnoreCase("memlag")) {
					if (sender instanceof Player) {
						Permission.SHOWTICKRATE.handle(sender);
						Permission.SHOWMEMORY.handle(sender);
						sender.sendMessage(ChatColor.GREEN + "[Static memory] " + ChatColor.YELLOW + "[Dynamic memory] " + ChatColor.RED + "[Unused memory]");
					}
					sender.sendMessage(PerformanceMonitor.getMemory(sender instanceof Player));
					sender.sendMessage(PerformanceMonitor.getTPS(sender instanceof Player));
				} else if (args[0].equalsIgnoreCase("ver") || args[0].equalsIgnoreCase("version")) {
					sender.sendMessage("NoLagg version " + this.getDescription().getVersion());
				} else if (args[0].equalsIgnoreCase("examine")) {
					int duration = 500;
				    try {
				    	 duration = Integer.parseInt(args[1]);
				    } catch (Exception ex) {}
					if (sender instanceof Player) {
						Permission.EXAMINE.handle(sender);
						PluginLogger.recipients.add(sender.getName());
					} else {
						PluginLogger.recipients.add(null);
					}
					if (PluginLogger.isRunning()) {
						CommonUtil.sendMessage(sender, ChatColor.RED + "The server is already being examined: " + PluginLogger.getDurPer() + "% completed");
						CommonUtil.sendMessage(sender, ChatColor.GREEN + "You will receive a notification when the report has been generated");
					} else {
						PluginLogger.duration = duration;
						PluginLogger.start();
						CommonUtil.sendMessage(sender, ChatColor.GREEN + "NoLagg will now monitor the server for " + duration + " ticks (" + (duration / 20) + " seconds)");
						CommonUtil.sendMessage(sender, ChatColor.GREEN + "You will receive a notification when the report has been generated");
					}
				} else {
					sender.sendMessage("Unknown sub-command!");
				}
			} else {
				Permission.SHOWTICKRATE.handle(sender);
				sender.sendMessage(PerformanceMonitor.getTPS(sender instanceof Player));
			}
		} catch (NoPermissionException ex) {
			sender.sendMessage(ChatColor.DARK_RED + "You don't have permission to use this command!");
		}
		return true;
	}

	@Override
	public void permissions() {
		this.loadPermissions(Permission.class);
	}

}
